/* This module implements features of xSEC (IBM proprietary power monitoring and capping.
  The source of this module is only intended to be divulged under NDA.
  IBM_AEM_IPMI_Interface_Design_v005.doc was the document used to produce this particular module.

  An example:
	ipmitool -I lanplus -U USERID -P PASSW0RD -H bmc2 raw 0x2e 0x80 0x4d 0x4f 0x00 0x00 0x0 00 0 01
  

  Clarifications from spec:
	-TODO: verify  It seems that the firmware version returns 2.1 if the version is 2.1 or newer.  This module will only support 2.1 or newer for the sake of simplicity.

  The api will be:
  	-issue request
	-if this layer must query more data from the IMM, it will populate a 'next-request' payload
	-If this layer populates the next-request pointer, it also will provide 'response-handler' and 'opaque_data'
	-provide a response_handler pointer and next-request payload
	-client will call r

  This only comments a quick-reference sampling of the specification.  This quick reference will omit system p and in depth explanation of what it describes.
  The netfn for all commands is 0x2e.  The IANA is 0x4d, 0x4f, 0x00. 
  Breaking it down, all functions are under the 0x2e netfn.  
  There are three commands to cover the specification:
	"Find Firmware (0x80)":  System x only: grab 4 'SDRR IDs and supported instances
		1-3: IANA
		4: reserved
		5-6: record id
		7-8: 00 01 (AME_MODULE_TYPE_ID)
	    -response:
		1: 00 if success
		2-4: IANA
		5: number of instances for module type
		6: reserved
		7-14: up to 4 16-bit ids
	"Invoke Element Method (0x81)": 
		1-3: IANA
		4: instance number (system x interpretation)
		5: element id
		6: method to invoke
		7-...: method arguments
	    response:
		1: 00 if successfull
		2-4: iana
		5-...: depends on method
	"Find Firmware instance (0x82)":
		1-3: IANA
		4: Instance Number
		5-6: 00 01

   The meat of productive work is done via 'invoking methods'.  A method can be thought of in the same way as an IPMI command, but implemented as another layer wrapped in a single command.  The methods are:
	read config 0x80
	read buffer 0x81
	read register 0x82 
	write register 0x83

   The methods supported in detail
	-Read element configuration
		0x2e (netfn, not bitshifted)
		0x81 (command: invoke element method
		0x4d IANA
		0x4f .
		0x00 .
		0xzz (instance number, assuming 0 for the moment)
		0xyy element id
		0x80 (read element config)
	  -response:
		0x00 (if success)

		0x00 .
		0xll length
		0x00 reserved
		..... depends on element


	A conversation to do something meaningful:
	-INITIALIZATION-
	0x2e 0x82 0x4d 0x4f 0x00 0 0 1 (read firmware info on the power meter)
		0 (should always be this for success)
		4d 4f 00 (should always be this for IBM systems)
		01 (if greater than one, multiple xSEC power meters exist, this example just has one)
		02 01 (xSEC revision number)
		20 (This number is the index to the power meter, must be stored for future reference)
		00 9b (Allegedly, ignore on rackmounts)
		
	
		
*/
#include "EnergyManager.h"
#define XSEC_FIND_FIRMWARE_REQUEST {0x2e,0x82,0x4d,0x4f,0x00,0,0,1};
EnergyManager * new_energymanager() {
	EnergyManager *em;
	em = (EnergyManager*)malloc(sizeof(EnergyManager));
	em->next_payload=0;
	em->data=0;
	em->data_units=0;
	em->data_label=0;
	em->dc_histogram_counts=0;
	em->snapshot_buffer=0;
	return em;
}

void init_energymanager(EnergyManager * em) { 
	if (em->next_payload != 0) { free(em->next_payload); em->next_payload=0; } /* shouldn't be possible, but just in case */
	em->next_payload=(unsigned char*)malloc(9);
	em->next_payload[0]=8; /* size of the request */
	em->next_payload[1]=0x2e; /* NetFn */
	em->next_payload[2]=0x82; /* Find Firmware Instance command */
	em->next_payload[3]=0x4d; /* IBM IANA */
	em->next_payload[4]=0x4f; /* IBM IANA */
	em->next_payload[5]=0x00; /* IBM IANA */
	em->next_payload[6]=0x00; /* Request instance 0 */
	em->next_payload[7]=0x00; /* module type of AME_MODULE_TYPE_ID */
	em->next_payload[8]=0x01; /* module type continued */
	em->next_response_handler=parse_init_data;
	return;
}

void parse_init_data(EnergyManager *em, unsigned char* data, int length) {
/* data is the IPMI payload received from the remote system, and length is the size of data */
/* parse output from a init request and return a struct to reflect context */
/* This code for now only supports a device that advertises 2.1 support for a single
   power meter, and the only context that is stored is the module handle */
	if (em->next_payload != 0) {  /* free memory consumed by the next_payload data */
		free(em->next_payload);
		em->next_payload=0;
	}
	em->next_response_handler=0;
	if (length < 10) {
		em->errorcode = ERROR_INVALID_RESPONSE;
		return;
	}
	if (data[5] != 2 || (data[6] != 1 && data[6] != 6 && data[6] != 9)) { /* only support something at 2.1 or 2.6 or 2.9 */
                em->errorcode = ERROR_UNSUPPORTED_REVISION;
		return;
	}
	if (data[4] != 1) { /* error for multiple power meters for now */
		em->errorcode = ERROR_UNSUPPORTED_FEATURE;
		return;
	}
	em->errorcode = 0;
	em->module_instance=data[7];
	return;
}

void prep_get_maxcap(EnergyManager * em) { /* get the minimum cap.  This is sufficient for system x support at the moment */
	if (em->next_payload != 0) {  /* free memory consumed by the next payload if already used */
		free(em->next_payload);
		em->next_payload=0;
	}
	em->next_payload=(unsigned char*)malloc(11);
	em->next_payload[0]=10; /* size of the request */
	em->next_payload[1]=0x2e; /* NetFn */
	em->next_payload[2]=0x81; /* Invoke element method */
	em->next_payload[3]=0x4d; /* IBM IANA */
	em->next_payload[4]=0x4f; /* IBM IANA */
	em->next_payload[5]=0x00; /* IBM IANA */
	em->next_payload[6]=em->module_instance; 
	em->next_payload[7]=0x07; /* The Power control element */
	em->next_payload[8]=0x82; /* read register method */
	em->next_payload[9]=0x02; /*  pcapmax */
	em->next_payload[10]=0x02; /* cap is 16 bits */
	if (em->data_label) { free(em->data_label); em->data_label=0; }
	em->data_label=strdup("cappingmax");
	if (em->data_units != 0) { free(em->data_units); em->data_units = 0; }
	em->data_units = strdup("mW");
	em->next_response_handler=parse_capping_data;
}
void prep_set_capenable(EnergyManager* em, int enable) {
	if (em->next_payload != 0) {  /* free memory consumed by the next payload if already used */
		free(em->next_payload);
		em->next_payload=0;
	}
	em->next_payload=(unsigned char*)malloc(12);
	em->next_payload[0]=12; /* size of the request */
	em->next_payload[1]=0x2e; /* NetFn */
	em->next_payload[2]=0x81; /* Invoke element method */
	em->next_payload[3]=0x4d; /* IBM IANA */
	em->next_payload[4]=0x4f; /* IBM IANA */
	em->next_payload[5]=0x00; /* IBM IANA */
	em->next_payload[6]=em->module_instance; 
	em->next_payload[7]=0x00; /* The status/control element */
	em->next_payload[8]=0x83; /* write register method */
	em->next_payload[9]=0x03; /* sms policy register */
	em->next_payload[10]=0x01; /* control register is 8 bits */
	if (enable==1) {
		em->next_payload[11]=2;
	} else {
		em->next_payload[11]=1;
	}
	em->next_response_handler=0;
}

void prep_set_cap(EnergyManager* em, int cap) {
	if (em->next_payload != 0) {  /* free memory consumed by the next payload if already used */
		free(em->next_payload);
		em->next_payload=0;
	}
	em->next_payload=(unsigned char*)malloc(13);
	em->next_payload[0]=12; /* size of the request */
	em->next_payload[1]=0x2e; /* NetFn */
	em->next_payload[2]=0x81; /* Invoke element method */
	em->next_payload[3]=0x4d; /* IBM IANA */
	em->next_payload[4]=0x4f; /* IBM IANA */
	em->next_payload[5]=0x00; /* IBM IANA */
	em->next_payload[6]=em->module_instance; 
	em->next_payload[7]=0x07; /* The Power control element */
	em->next_payload[8]=0x83; /* read register method */
	em->next_payload[9]=0x00; /* cap register */
	em->next_payload[10]=0x02; /* cap is 16 bits */
	cap = cap/100; /* divide to count of 100 mw */
	cap &= 0xffff; /* mask any leading bits, just in case */
	em->next_payload[11]=((cap&0xff00)>>8); /* copy msb */
	em->next_payload[12]=(cap&0xff); /* lsb */
	em->next_response_handler=0;
}
void prep_get_powerstatus(EnergyManager * em) { /* get the minimum cap.  This is sufficient for system x support at the moment */
	if (em->next_payload != 0) {  /* free memory consumed by the next payload if already used */
		free(em->next_payload);
		em->next_payload=0;
	}
	em->next_payload=(unsigned char*)malloc(11);
	em->next_payload[0]=10; /* size of the request */
	em->next_payload[1]=0x2e; /* NetFn */
	em->next_payload[2]=0x81; /* Invoke element method */
	em->next_payload[3]=0x4d; /* IBM IANA */
	em->next_payload[4]=0x4f; /* IBM IANA */
	em->next_payload[5]=0x00; /* IBM IANA */
	em->next_payload[6]=em->module_instance; 
	em->next_payload[7]=0x00; /* The power status/control element */
	em->next_payload[8]=0x82; /* read register method */
	em->next_payload[9]=0x00; /* status register */
	em->next_payload[10]=0x01; /* 8 bits */
	em->next_response_handler=parse_status_data;
}
void parse_status_data(EnergyManager *em, unsigned char* data, int length) {
	int decodestatus;
	if (em->next_payload != 0) {  /* free memory consumed by the next_payload data */
		free(em->next_payload);
		em->next_payload=0;
	}
	if (length <  5) {
		em->errorcode=ERROR_INVALID_RESPONSE;
		return;
	}
	if (data[0] != 0) { 
		em->errorcode=ERROR_INVALID_RESPONSE;
	}
	/* check iana TODO */
	decodestatus = data[4] >> 1;
	decodestatus &= 7;
	if (decodestatus == 2 || decodestatus == 5) {
		em->pcap_enabled=1;
	} else {
		em->pcap_enabled=0;
	}
}

void prep_get_snapshot_buffer(EnergyManager * em) { /* get the snapshot buffer data */
	if (em->next_payload != 0) {  /* free memory consumed by the next payload if already used */
		free(em->next_payload);
		em->next_payload=0;
	}
	/* in this round, we must simply retrieve the total size and size-at-a-time values */
	em->next_payload=(unsigned char*)malloc(9);
	em->next_payload[0]=12; /* size of the request */
	em->next_payload[1]=0x2e; /* NetFn */
	em->next_payload[2]=0x81; /* Invoke element method  */
	em->next_payload[3]=0x4d; /* IBM IANA */
	em->next_payload[4]=0x4f; /* IBM IANA */
	em->next_payload[5]=0x00; /* IBM IANA */
	em->next_payload[6]=em->module_instance; 
	em->next_payload[7]=0x03; /* Snapshot buffer */
	em->next_payload[8]=0x81; /* read snapshot */
	em->next_payload[9]=em->dc_buffer_offset>>8; /* offset */
	em->next_payload[10]=em->dc_buffer_offset&0xff;
	if (em->snapshot_required_fetches > 1) { /* if this is not last packet, use largest supported packet size */
		em->next_payload[11]=em->snapshot_buffer_chunk>>8;
		em->next_payload[12]=em->snapshot_buffer_chunk&0xff;
	} else { /* last packet, only need remaining data */
        	em->next_payload[11] = em->snapshot_remainder_size>>8; 
        	em->next_payload[12] = em->snapshot_remainder_size&0xff; 
	}
	em->next_response_handler=parse_snapbuffer;
}

void parse_snapbuffer(EnergyManager* em, unsigned char* data, int length) { /*  */
	int buf_tag;
	int num_bytes;
	int remaining_bytes;
	int histogram_offset;
	int hist_idx;
	int val_idx;
	unsigned char* curptr;
	curptr=data;
	curptr += 8;
	if (em->next_payload != 0) {  /* free memory consumed by the next_payload data */
		free(em->next_payload);
		em->next_payload=0;
	}
	if (length <  10) {
		em->errorcode=ERROR_INVALID_RESPONSE;
		return;
	}
	if (data[0] != 0) { 
		em->errorcode=ERROR_INVALID_RESPONSE;
		return;
	}
	/* check iana TODO */
	buf_tag=(data[4]<<8)+data[5];
	if (em->firstbuffer) {  /* first packet, store tag for comparison */ 
		em->current_buffer_tag=buf_tag;
		em->firstbuffer=0;
	} else if (buf_tag != em->current_buffer_tag) { /* must start over, data invalid */
		em->snapshot_required_fetches = em->snapshot_total_fetches;
		em->dc_buffer_offset=0;
		em->firstbuffer=1;
                prep_get_snapshot_buffer(em);
		return;
	}
	/* we have data coming in to put into the histogram (the only part of the snapshot buffer that we 
	 * curreontly care about */
        num_bytes = (data[6]>>8)+(data[7]);
	remaining_bytes = num_bytes;
	histogram_offset=(em->dc_buffer_offset)-9;
	if (histogram_offset < 0) {
		remaining_bytes +=histogram_offset; /* mark that offset number of bytes are discarded */
		curptr=curptr-histogram_offset; /* advance pointer to start of histogram buckets */
		histogram_offset=0;
	}
	hist_idx=histogram_offset/em->dc_histogram_binsize;
	while (remaining_bytes > 0 && hist_idx < em->dc_histogram_bincount) {
		val_idx=em->dc_histogram_binsize-(histogram_offset%em->dc_histogram_binsize)-1;
		if (val_idx==(em->dc_histogram_binsize-1)) { /* new value */
			em->dc_histogram_counts[hist_idx] = curptr[0]<<(8*val_idx);
		} else {
			em->dc_histogram_counts[hist_idx] += curptr[0]<<(8*val_idx);
		}
		curptr+=1;
		histogram_offset+=1;
		remaining_bytes-=1;
		hist_idx=histogram_offset/em->dc_histogram_binsize;
	}
/* variables available to me here
	em->dc_histogram_bincount = data[10];
	em->dc_histogram_binrange = data[11];
	em->dc_histogram_binsize = data[12];
*/
	em->snapshot_required_fetches -= 1; /* decrease remaining fetches */
	em->dc_buffer_offset+=num_bytes; /* advance offset into snapshot buffer */
	if (em->snapshot_required_fetches > 0) {
		prep_get_snapshot_buffer(em);
	}
}

void prep_get_snapshot_buffer_config(EnergyManager * em) { /* get the snapshot buffer data */
	if (em->next_payload != 0) {  /* free memory consumed by the next payload if already used */
		free(em->next_payload);
		em->next_payload=0;
	}
	/* in this round, we must simply retrieve the total size and size-at-a-time values */
	em->next_payload=(unsigned char*)malloc(9);
	em->next_payload[0]=8; /* size of the request */
	em->next_payload[1]=0x2e; /* NetFn */
	em->next_payload[2]=0x81; /* Invoke element method  */
	em->next_payload[3]=0x4d; /* IBM IANA */
	em->next_payload[4]=0x4f; /* IBM IANA */
	em->next_payload[5]=0x00; /* IBM IANA */
	em->next_payload[6]=em->module_instance; 
	em->next_payload[7]=0x03; /* Snapshot buffer */
	em->next_payload[8]=0x80; /* read snapshot config */
	em->next_response_handler=parse_buffer_config;
}

void prep_get_dc_histogram_config(EnergyManager * em) { /* get the snapshot buffer data */
	if (em->next_payload != 0) {  /* free memory consumed by the next payload if already used */
		free(em->next_payload);
		em->next_payload=0;
	}
	/* in this round, we must simply retrieve the total size and size-at-a-time values */
	em->next_payload=(unsigned char*)malloc(9);
	em->next_payload[0]=8; /* size of the request */
	em->next_payload[1]=0x2e; /* NetFn */
	em->next_payload[2]=0x81; /* Invoke element method  */
	em->next_payload[3]=0x4d; /* IBM IANA */
	em->next_payload[4]=0x4f; /* IBM IANA */
	em->next_payload[5]=0x00; /* IBM IANA */
	em->next_payload[6]=em->module_instance; 
	em->next_payload[7]=0x02; /* dc histogram element */
	em->next_payload[8]=0x80; /* read histogram config */
	em->next_response_handler=parse_histogram_config;
}
void parse_histogram_config(EnergyManager* em, unsigned char* data, int length) { /* extract histogram specifications */
	if (em->next_payload != 0) {  /* free memory consumed by the next_payload data */
		free(em->next_payload);
		em->next_payload=0;
	}
	if (length <  10) {
		em->errorcode=ERROR_INVALID_RESPONSE;
		return;
	}
	if (data[0] != 0) { 
		em->errorcode=ERROR_INVALID_RESPONSE;
		return;
	}
	/* check iana TODO */
	em->dc_histogram_smallest_bin = (data[6] << 8) + data[7];
	em->dc_histogram_bincount = data[10];
	em->dc_histogram_binrange = data[11];
	em->dc_histogram_binsize = data[12];
	em->dc_histogram_accuracy = data[13]/2.0;
	if (em->dc_histogram_counts) {
		free(em->dc_histogram_counts);
		em->dc_histogram_counts=0;
	}
	em->dc_histogram_counts = (int*)malloc(em->dc_histogram_bincount*sizeof(int));
	if (em->dc_fetch_bufconfig) {
		prep_get_snapshot_buffer_config(em);
	}
}
void parse_buffer_config(EnergyManager* em, unsigned char* data, int length) { /* extract interesting accuracy figures from the EAR */
	if (em->next_payload != 0) {  /* free memory consumed by the next_payload data */
		free(em->next_payload);
		em->next_payload=0;
	}
	if (length <  10) {
		em->errorcode=ERROR_INVALID_RESPONSE;
		return;
	}
	if (data[0] != 0) { 
		em->errorcode=ERROR_INVALID_RESPONSE;
		return;
	}
	/* check iana TODO */
	em->snapshot_buffer_chunk = (data[6] << 8) + data[7];
	em->snapshot_buffer_chunk -= 12; /* bug when traffic encrypted, workaround */
	em->snapshot_buffer_size = (data[10] << 8) + data[11];
	em->snapshot_required_fetches = em->snapshot_buffer_size/em->snapshot_buffer_chunk;
	em->snapshot_remainder_size = em->snapshot_buffer_size % em->snapshot_buffer_chunk;
	if (em->snapshot_remainder_size) {
		em->snapshot_required_fetches += 1;
	}
	em->snapshot_total_fetches = em->snapshot_required_fetches;
	em->dc_buffer_offset=0;
	em->firstbuffer=1;
	if (em->dc_fetch_snapshotbuf) {
                prep_get_snapshot_buffer(em);
	}
}
void prep_get_cap(EnergyManager * em) { /* get the minimum cap.  This is sufficient for system x support at the moment */
	if (em->next_payload != 0) {  /* free memory consumed by the next payload if already used */
		free(em->next_payload);
		em->next_payload=0;
	}
	em->next_payload=(unsigned char*)malloc(11);
	em->next_payload[0]=10; /* size of the request */
	em->next_payload[1]=0x2e; /* NetFn */
	em->next_payload[2]=0x81; /* Invoke element method */
	em->next_payload[3]=0x4d; /* IBM IANA */
	em->next_payload[4]=0x4f; /* IBM IANA */
	em->next_payload[5]=0x00; /* IBM IANA */
	em->next_payload[6]=em->module_instance; 
	em->next_payload[7]=0x07; /* The Power control element */
	em->next_payload[8]=0x82; /* read register method */
	em->next_payload[9]=0x00; /* EARac register */
	em->next_payload[10]=0x02; /* cap is 16 bits */
	if (em->data_label) { free(em->data_label); em->data_label=0; }
	em->data_label=strdup("cappingvalue");
	if (em->data_units != 0) { free(em->data_units); em->data_units = 0; }
	em->data_units = strdup("mW");
	em->next_response_handler=parse_capping_data;
}
void prep_get_mincap(EnergyManager * em) { /* get the minimum cap.  This is sufficient for system x support at the moment */
	if (em->next_payload != 0) {  /* free memory consumed by the next payload if already used */
		free(em->next_payload);
		em->next_payload=0;
	}
	em->next_payload=(unsigned char*)malloc(11);
	em->next_payload[0]=10; /* size of the request */
	em->next_payload[1]=0x2e; /* NetFn */
	em->next_payload[2]=0x81; /* Invoke element method */
	em->next_payload[3]=0x4d; /* IBM IANA */
	em->next_payload[4]=0x4f; /* IBM IANA */
	em->next_payload[5]=0x00; /* IBM IANA */
	em->next_payload[6]=em->module_instance; 
	em->next_payload[7]=0x07; /* The Power control element */
	em->next_payload[8]=0x82; /* read register method */
	em->next_payload[9]=0x04; /* EARac register */
	em->next_payload[10]=0x02; /* cap is 16 bits */
	if (em->data_label) { free(em->data_label); em->data_label=0; }
	em->data_label=strdup("cappingmin");
	if (em->data_units != 0) { free(em->data_units); em->data_units = 0; }
	em->data_units = strdup("mW");
	em->next_response_handler=parse_capping_data;
}
	
void parse_capping_data(EnergyManager * em, unsigned char* data, int length) {
	int capamount;
	int i;
	int shift;
	if (em->next_payload != 0) {  /* free memory consumed by the next_payload data */
		free(em->next_payload);
		em->next_payload=0;
	}
	if (length <  6) {
		em->errorcode=ERROR_INVALID_RESPONSE;
		return;
	}
	if (data[0] != 0) { 
		em->errorcode=ERROR_INVALID_RESPONSE;
	}
	/* check iana TODO */
	if (em->data != 0) {
		free(em->data);
		em->data=0;
	}
	em->data_length = sizeof(int);
	capamount=((data[4]<<8)+data[5])*100;
	em->data = (unsigned char*)malloc(sizeof(int));
	shift = 8*(sizeof(int)-1);
	for (i=0;i<sizeof(int);i++) {
		em->data[i] =(capamount&(0xff<<shift))>>shift;
		shift-=8;
	}
}

void prep_get_precision(EnergyManager * em) { /* get the AC millijoule amount */
	if (em->next_payload != 0) {  /* free memory consumed by the next payload if already used */
		free(em->next_payload);
		em->next_payload=0;
	}
	em->next_payload=(unsigned char*)malloc(9);
	em->next_payload[0]=10; /* size of the request */
	em->next_payload[1]=0x2e; /* NetFn */
	em->next_payload[2]=0x81; /* Invoke element method */
	em->next_payload[3]=0x4d; /* IBM IANA */
	em->next_payload[4]=0x4f; /* IBM IANA */
	em->next_payload[5]=0x00; /* IBM IANA */
	em->next_payload[6]=em->module_instance; 
	em->next_payload[7]=0x01; /* The EAR Element */
	em->next_payload[8]=0x80; /* read config */
	em->next_response_handler=parse_ear_cfgdata;
}
void prep_get_ac_energy(EnergyManager * em) { /* get the AC millijoule amount */
	if (em->next_payload != 0) {  /* free memory consumed by the next payload if already used */
		free(em->next_payload);
		em->next_payload=0;
	}
	em->next_payload=(unsigned char*)malloc(11);
	em->next_payload[0]=10; /* size of the request */
	em->next_payload[1]=0x2e; /* NetFn */
	em->next_payload[2]=0x81; /* Invoke element method */
	em->next_payload[3]=0x4d; /* IBM IANA */
	em->next_payload[4]=0x4f; /* IBM IANA */
	em->next_payload[5]=0x00; /* IBM IANA */
	em->next_payload[6]=em->module_instance; 
	em->next_payload[7]=0x01; /* The EAR Element */
	em->next_payload[8]=0x82; /* read register method */
	em->next_payload[9]=0x01; /* EARac register */
	em->next_payload[10]=0x08; /* EARac is 64 bits */
	if (em->data_label) {
		free(em->data_label);
		em->data_label=0;
	}
	em->data_label=strdup("AC Energy Usage");
	em->next_response_handler=parse_ear_data;
}

void prep_get_dc_energy(EnergyManager * em) { /* get the AC millijoule amount */
	if (em->next_payload != 0) {  /* free memory consumed by the next payload if already used */
		free(em->next_payload);
		em->next_payload=0;
	}
	em->next_payload=(unsigned char*)malloc(11);
	em->next_payload[0]=10; /* size of the request */
	em->next_payload[1]=0x2e; /* NetFn */
	em->next_payload[2]=0x81; /* Invoke element method */
	em->next_payload[3]=0x4d; /* IBM IANA */
	em->next_payload[4]=0x4f; /* IBM IANA */
	em->next_payload[5]=0x00; /* IBM IANA */
	em->next_payload[6]=em->module_instance; 
	em->next_payload[7]=0x01; /* The EAR Element */
	em->next_payload[8]=0x82; /* read register method */
	em->next_payload[9]=0x00; /* EARdc register */
	em->next_payload[10]=0x08; /* EARac is 64 bits */
	em->data_label=strdup("DC Energy Usage");
	em->next_response_handler=parse_ear_data;
}

void parse_ear_cfgdata(EnergyManager* em, unsigned char* data, int length) { /* extract interesting accuracy figures from the EAR */
	if (em->next_payload != 0) {  /* free memory consumed by the next_payload data */
		free(em->next_payload);
		em->next_payload=0;
	}
	if (length <  10) {
		em->errorcode=ERROR_INVALID_RESPONSE;
		return;
	}
	if (data[0] != 0) { 
		em->errorcode=ERROR_INVALID_RESPONSE;
	}
	/* check iana TODO */
	em->ear_dc_precision = data[8]*5;
	em->ear_ac_precision = data[9]*5;
}

void parse_ear_data(EnergyManager * em, unsigned char* data, int length) {
	if (em->next_payload != 0) {  /* free memory consumed by the next_payload data */
		free(em->next_payload);
		em->next_payload=0;
	}
	if (length <  12) {
		em->errorcode=ERROR_INVALID_RESPONSE;
		return;
	}
	if (data[0] != 0) { 
		em->errorcode=ERROR_INVALID_RESPONSE;
	}
	/* check iana TODO */
	if (em->data != 0) {
		free(em->data);
		em->data=0;
	}
	em->data = (unsigned char*)malloc(8);
	em->data_length = 8;
	em->data[0] = data[4];
	em->data[1] = data[5];
	em->data[2] = data[6];
	em->data[3] = data[7];
	em->data[4] = data[8];
	em->data[5] = data[9];
	em->data[6] = data[10];
	em->data[7] = data[11];
	if (em->data_units != 0) { free(em->data_units); em->data_units = 0; }
	em->data_units = strdup("mJ");
}
	
	


////////unsigned char* get_em_buffer_info_request(em_context ctx) { 
////////	unsigned char* request_data;
////////	request_data=(unsigned char*)malloc(9);
////////	request_data[0]=8; /* size of the request */
////////	request_data[1]=0x2e; /* NetFn */
////////	request_data[2]=0x81; /* Invoke element method */
////////	request_data[3]=0x4d; /* IBM IANA */
////////	request_data[4]=0x4f; /* IBM IANA */
////////	request_data[5]=0x00; /* IBM IANA */
////////	request_data[6]=ctx.module_instance; /* The specific instance to query */
////////	request_data[7]=0x03; /* the snapshot buffer element id */
////////	request_data[8]=0x80; /* read config data */
////////	return request_data;
////////}


	
    /*
int main(int argc, char* argv[]) {
	int i;
	int size;
	EnergyManager *em = new_powermanager();
	init_powermanager(em);
	for (i=1;i<=em->next_payload[0];i++) {
		printf("0x%02x ",em->next_payload[i]);
	}
	printf("\n");
} */
